
# Mash constants
MAX_MASH_DIST = '0.1'
UNFILTERED_MASH_DIST = '0.3'

# DNA fragmentation constants
FRAGMENT_SIZE = 1020
MIN_FRAGMENT_SIZE = 100

CITATION = '''Schwengers O., Hain T., Chakraborty T., Goesmann A. (2019)
ReferenceSeeker: rapid determination of appropriate reference genomes.
bioRxiv 863621; doi: https://doi.org/10.1101/863621'''
